import time
import json
from enum import Enum, auto

import meraki

from .config import *
from .exceptions import *

__version__ = '0.1.0a'


class BatchHelperStatus(Enum):
    """Batch helper statuses."""

    COMPLETE = auto()
    PREPARED = auto()
    WORKING = auto()
    PENDING = auto()
    FAILED = auto()


class BatchHelper:

    def __init__(self,
                 dashboard_session: meraki.DashboardAPI,
                 organizationId: str,
                 new_actions: list,
                 dependent: bool = False,
                 required_batch_id: str = "__default__",
                 alternateOrganizationId: str = "__default__",
                 linear_new_batches: bool = False,
                 confirmed_new_batches: bool = True,
                 synchronous_new_batches: bool = False,
                 actions_per_new_batch: int = MAX_ACTIONS_ASYNC,
                 interval_factor: float = MINIMUM_INTERVAL_FACTOR,
                 maximum_wait: int = MAXIMUM_WAIT):
        """ Creates the batch helper.
        @param dashboard_session: the Meraki Python SDK dashboard session used for executing the batch
        @param organizationId: the Meraki organization ID of the relevant organization where the batch will be run
        @param new_actions: the list of all aggregated new actions that the batch helper will parse
        @param dependent: whether the new actions rely on a different batch to finish first
        @param required_batch_id: when dependent, required_batch_id is the ID of the batch upon which the new batches
        depend
        @param alternateOrganizationId: supplied if the required batch is from a different organization ID
        @param linear_new_batches: whether the new actions need to be executed linearly; requires confirmed new batches
        @param confirmed_new_batches: whether the submitted batches will start upon submission
        @param synchronous_new_batches: whether the submitted batches will be synchronous
        @param actions_per_new_batch: maximum number of actions to group into a single batch
        @param interval_factor: the factor to use for determining wait time
        @param maximum_wait: the maximum wait time in seconds for the entire list of new_actions
        """

        # self assignments
        self.dashboard_session = dashboard_session
        self.organizationId = organizationId
        self.new_actions = new_actions
        self.dependent = dependent
        self.required_batch_id = required_batch_id
        self.alternateOrganizationId = alternateOrganizationId
        self.linear_new_batches = linear_new_batches
        self.confirmed_new_batches = confirmed_new_batches
        self.synchronous_new_batches = synchronous_new_batches
        self.actions_per_new_batch = actions_per_new_batch
        self.interval_factor = interval_factor
        self.maximum_wait = maximum_wait
        self.status = BatchHelperStatus.PENDING

        # defaults prior to preparation
        self.preview = ''
        self.new_batches = list()
        self.new_batches_responses = list()
        self.submitted_new_batches_ids = list()
        self.successful_new_batch_ids = list()
        self.failed_new_batch_ids = list()

        # Validate batch helper input against common mistakes.
        print(f'Validating input...')

        # validate number of actions per batch
        if self.actions_per_new_batch > 100:
            status = BatchHelperStatus.FAILED
            raise TooManyActionsError(self.actions_per_new_batch)

        if self.actions_per_new_batch < 2:
            status = BatchHelperStatus.FAILED
            raise NotEnoughActionsError(self.actions_per_new_batch)

        # validate number of actions per batch in an async batch
        if self.actions_per_new_batch > 20 and self.synchronous_new_batches:
            status = BatchHelperStatus.FAILED
            raise TooManySynchronousActionsError(self.actions_per_new_batch)

        # validate polling interval
        if self.interval_factor < MINIMUM_INTERVAL_FACTOR:
            status = BatchHelperStatus.FAILED
            raise IntervalFactorTooSmallError(interval_factor)

        # validate required batch ID
        if self.dependent and self.required_batch_id == "__default__":
            status = BatchHelperStatus.FAILED
            raise RequiredBatchIdError()

        # enforce linear new batches requirements
        if self.linear_new_batches and not self.confirmed_new_batches:
            status = BatchHelperStatus.FAILED
            raise LinearBatchRequirementsError()

        if self.alternateOrganizationId == '__default__':
            self.required_batch_org_id = self.organizationId
        else:
            self.required_batch_org_id = self.alternateOrganizationId

    def group_actions(self):
        """ Groups actions into lists of appropriate size and returns a list generator. """
        total_actions = len(self.new_actions)
        for i in range(0, total_actions, self.actions_per_new_batch):
            yield self.new_actions[i:i + self.actions_per_new_batch]

    def prepare(self):
        """ Groups actions into batches of the appropriate size. """
        grouped_actions_list = list(self.group_actions())
        created_batches = list()

        # Add each new batch to the new_batches list
        for action_list in grouped_actions_list:
            batch = {
                "organizationId": self.organizationId,
                "actions": action_list,
                "synchronous": self.synchronous_new_batches,
                "confirmed": self.confirmed_new_batches
            }
            created_batches.append(batch)

        self.new_batches = created_batches
        self.status = BatchHelperStatus.PREPARED

    def wait_for_required_batch(self):
        """ When the batches are dependent or linear, waits for the required batch to finish. """

        time_waited = 0
        completed = False
        while time_waited < self.maximum_wait and not completed:
            # While we haven't waited the maximum time and the required batch hasn't completed
            try:
                # Check on the required batch
                required_batch = self.dashboard_session.organizations.getOrganizationActionBatch(
                    self.required_batch_org_id, self.required_batch_id
                )
            except meraki.APIError:
                # If it doesn't exist, raise error
                status = BatchHelperStatus.FAILED
                raise RequiredBatchNotFoundError(self.required_batch_id, self.required_batch_org_id)

            if required_batch['status']['completed']:
                # If it's complete, stop waiting
                self.successful_new_batch_ids.append(required_batch['id'])
                completed = True

            if not required_batch['confirmed']:
                # If it's not confirmed, then it hasn't started, so stop waiting.
                self.status = BatchHelperStatus.FAILED
                raise RequiredBatchNotStartedError(self.required_batch_id, self.required_batch_org_id)

            # if it's started, but it's not complete, check its status on an interval
            if required_batch['confirmed'] and not required_batch['status']['completed']:
                if required_batch['status']['failed']:
                    # if it failed, then stop waiting
                    self.status = BatchHelperStatus.FAILED
                    self.failed_new_batch_ids.append(required_batch['id'])

                    raise RequiredBatchFailureError(self.required_batch_id,
                                                    batch_errors=required_batch['status']['errors'])

                # wait a reasonable interval dependent upon the number of actions in the batch
                length_of_required_batch_actions = len(required_batch['actions'])
                interval = length_of_required_batch_actions * self.interval_factor

                # don't wait longer than the maximum
                if time_waited + interval > MAXIMUM_WAIT:
                    interval = MAXIMUM_WAIT - time_waited - 1

                # update elapsed timer to ensure max wait is not exceeded
                time_waited += interval

                # Print update to console and start waiting
                print(f'Required batch ID {self.required_batch_id} in progress. Waiting {interval} seconds for it to '
                      f'complete.')

                # wait
                time.sleep(interval)

        if time_waited >= self.maximum_wait:
            # error if it's taken too long
            status = BatchHelperStatus.FAILED
            raise RequiredBatchStillInProgress(self.required_batch_id, self.required_batch_org_id)

        if completed:
            # return True if completed
            return True

    def check_batch_queue(self):
        """ Returns the organization's batch queue: pending, active, and whether full. """
        pending_action_batches = self.dashboard_session.organizations.getOrganizationActionBatches(
            organizationId=self.organizationId,
            status='pending')
        active_action_batches = [batch for batch in pending_action_batches if batch['confirmed']]

        # Dashboard API supports up to MAXIMUM_ACTIVE_ACTION_BATCHES.
        batch_queue_is_full = True if len(active_action_batches) >= MAXIMUM_ACTIVE_ACTION_BATCHES else False

        print(f'Checking batch queue...')
        return pending_action_batches, active_action_batches, batch_queue_is_full

    def find_batch_queue_capacity(self):
        """ Finds capacity on the batch queue. """
        pending_action_batches, active_action_batches, batch_queue_is_full = self.check_batch_queue()
        number_of_active_batches = len(active_action_batches)
        print(f'There are {number_of_active_batches} active action batches.')

        if len(active_action_batches):
            # If there are any active action batches, then flatten the list of actions; we want to know how many total
            # actions are pending, so we can calculate a reasonable wait time. Waiting too long is no fun, but
            # checking too frequently is inefficient, and might impact other API applications.
            print(f'Finding batch queue capacity...')

            # Gather all the action lists for the existing action batches
            active_batch_action_lists = [batch['actions'] for batch in active_action_batches]

            # Flatten the list of action lists into a single list of actions
            active_batch_actions = [action for action_list in active_batch_action_lists for action in action_list]

            # Calculate the mean number of actions per batch
            mean_actions_per_batch = len(active_batch_actions) / number_of_active_batches

            # Set an interval based on the interval_factor and the mean actions per batch
            interval = self.interval_factor * mean_actions_per_batch

            while batch_queue_is_full:
                # Wait for space in the queue until there's an open slot.
                print(f'There are already {number_of_active_batches} active action batches. Waiting {interval} '
                      f'seconds before trying again.')
                time.sleep(interval)

                active_action_batches, active_action_batches, batch_queue_is_full = self.check_batch_queue()
        return True

    def confirm_readiness_for_new_batch(self):
        """ Confirm that the org is ready to accept a new batch. """
        if self.dependent:
            # if dependent, check that the required batch has completed
            self.wait_for_required_batch()

        return True

    def submit_action_batches(self):
        """ Submit the next batch and remove it from the list of remaining batches. """
        try:
            new_batch_response = self.dashboard_session.organizations.createOrganizationActionBatch(**self.new_batches.pop(0))
        except meraki.APIError:
            self.status = BatchHelperStatus.FAILED
            raise BatchCreationFailureError()

        self.new_batches_responses.append(new_batch_response)
        self.submitted_new_batches_ids.append(new_batch_response['id'])

        print(f'Creating the next action batch. {len(self.new_batches)} action batches remain.')

        if not self.linear_new_batches:
            # set dependent false, and required batch ID to default
            self.dependent = False
            self.required_batch_id = '__default__'
        else:
            self.dependent = True
            self.required_batch_id = new_batch_response['id']

    def generate_preview(self):
        """ Generates a JSON preview of the new batches. """
        self.prepare()

        preview_json = json.dumps(self.new_batches, indent=2)
        with open('batch_helper_preview.json', 'w+', encoding='utf-8', newline='') as preview_file:
            preview_file.write(preview_json)

    def execute(self):
        """ Submits new batches. """
        while len(self.new_batches):
            self.status = BatchHelperStatus.WORKING
            # Loop as long as there are batches left to process
            print(f'Confirming readiness for batch submission...')

            # Check that the environment is ready for the new batch. Set a single variable false until it passes all the
            # while loop gates
            ready_for_new_batch = False
            while not ready_for_new_batch:
                # Confirm readiness before proceeding. Need to figure out how to pass this required batch ID in such a
                # way that it can change from the initial required batch, to the linear preceding batch, if linear
                ready_for_new_batch = self.confirm_readiness_for_new_batch()

            ready_for_new_batch = False
            while not ready_for_new_batch:
                ready_for_new_batch = self.find_batch_queue_capacity()

            print(f'Submitting new batch.')
            # submit new batch(es)
            self.submit_action_batches()

        self.status = BatchHelperStatus.COMPLETE

        return True




import re
from collections import Counter
from mc_user_info import *


def SplitCheckSerials(user_text,search_type):
    """
    This function will accept of user's input or command line parms and find the serial
    numbers and split them out to a list to return.
    :param user_text: The incoming user text from Teams or the command line
    :param search_type: 'Claim' or 'Translate' based on the command line type to parse
    :return: A list of Meraki serial numbers and a string if there was an issue
    """
    debug = DEBUG or DEBUG_SPLIT
    maybe_serials = list()
    regex = re.compile(r"\s*to\s *", flags=re.I)
    if search_type == "Claim":
        the_rest = regex.split(user_text)[0]
        the_rest = the_rest.strip()
        if debug:
            print(f"the_rest = {the_rest}")
        if debug:
            print(f"len(the_rest) = {len(the_rest)}")
        if len(the_rest) > 14:
            regex2 = re.compile(r"claim\s *", flags=re.I)
            maybe_serials = regex2.split(the_rest)[1]
    elif search_type == "Translate":
        maybe_serials = regex.split(user_text)[1]
    if not re.search(',', maybe_serials, re.IGNORECASE) == None:
        maybe_serials = maybe_serials.replace(" ", "")
    if debug:
        print(f"maybe_serials = {maybe_serials}")
        
    if len(maybe_serials)==0:
        return([],"I'm sorry, but {} is not a list of Meraki serial numbers delimited by commas, spaces or semicolons.".format(the_rest))
        
    serials = re.split(';|,|\s',maybe_serials)
    
    if debug:
        print(f"serials = {serials}")
    x = 0
    while x <= len(serials)-1:
        if debug:
            print(f"serials[{x}] = {serials[x]}")
            print(f"len(serials[{x}]) = {len(serials[x])}")
        if not len(serials[x]) == 14:
            return([],"I'm sorry, but {} is not a list of Meraki serial numbers delimited by commas, spaces or semicolons.".format(maybe_serials))
        x+=1
    duplicated_serials = [k for k, v in Counter(serials).items() if v > 1]
    if not len(duplicated_serials) == 0:
        return([],"I'm sorry, but {} cannot be repeated in the list of Meraki serial numbers.".format(', '.join(duplicated_serials)))
    return(serials,"")



from netmiko import ConnectHandler
from mc_user_info import DEBUG
import os


def GetConfig(host_id, ios_username, ios_password, ios_port, ios_secret):
    """
    This function will write a Catalyst switch config to a file.
    :param host: The hostname, fqdn or IP address of a switch/stack.
    :return: The hostname, the filespec for the saved config file.
    """

    debug = DEBUG

    Features_configured = list()
    aux_features_config = list()
    host_name = ""

    # SSH to the switch with netmiko, read the config, grab the
    # hostname, write the config out to a file using the hostname
    # as part of the filespec
    session_info = {
        'device_type': 'cisco_xe',
        'host': host_id,
        'username': ios_username,
        'password': ios_password,
        'port': ios_port,          # optional, defaults to 22
        'secret': ios_secret,      # optional, defaults to ''
    }
    net_connect = ConnectHandler(**session_info)
    switch_name = net_connect.find_prompt()
    net_connect.enable()
    switch_name = net_connect.find_prompt()
    switch_name = switch_name[:len(switch_name) - 1]
    net_connect.send_command('term len 0')
    config = net_connect.send_command('show running-config')
    net_connect.send_command('term len 24')
    net_connect.disconnect()
    dir = os.path.join(os.getcwd(), "../../files")
    config_file = os.path.join(dir, switch_name + ".cfg")
    file = open(config_file, "w")
    file.writelines(config)
    file.close()
    return(switch_name, config_file)



import batch_helper
import json
import pprint
import re
from ciscoconfparse2 import CiscoConfParse
from collections import defaultdict
from mc_user_info import DEBUG, DEBUG_TRANSLATOR
from mc_pedia import mc_pedia, nm_dict


def Evaluate(config_file, nm_list):
    """
    This parent function will evaluate a Catalyst switch config file and
    reports on which features that are being used can and cannot be mapped to
    Meraki features.
    :param config_file: The catalyst IOSXE config file to parse through
    :param nm_list: List of Network Modules per switch
    :return: Lists of uplinks, downlinks and other ports as well as a
    :      : dictionary with all ports and their settings, and the hostname
    """

    debug = DEBUG or DEBUG_TRANSLATOR

    port_dict = {}
    switch_dict = {}
    # List of interfaces that are shut
    shut_interfaces = list()

    def read_Cisco_SW():
        """
        This sub-function parses the Catalyst switch config file and report on
        which features that are being used can and cannot be mapped to Meraki
        features.
        :param NONE: Using global variables in the parent function
        :return: Lists of uplinks, downlinks and other ports as well as a
        :      : dictionary with all the ports and their settings, and the
        :      : hostname
        """
        for key, val in mc_pedia['switch'].items():
            switch_dict[key] = ""
        if debug:
            print(f"switch_dict = {switch_dict}")
        # Parsing the Cisco Catalyst configuration
        if debug:
            print("-------- Reading <"+config_file+"> Configuration --------")
        parse = CiscoConfParse(config_file, syntax='ios', factory=True)
        # Try out our mc_pedia for the switch name
        for key, val in mc_pedia['switch'].items():
            newvals = {}
            exec(val.get('iosxe'), locals(), newvals)
            switch_dict[key] = newvals[key]
            if debug:
                print(f"switch_dict['{key}'] = {switch_dict[key]}")

        Gig_uplink = list()
        Ten_Gig_uplink = list()
        Twenty_Five_Gig_uplink = list()
        Forty_Gig_uplink = list()
        Hundred_Gig_uplink = list()
        All_interfaces = defaultdict(list)

        # Grab all shutdown interfaces
        intf_for_test = parse.find_parent_objects([
            r'^interface',
            r'^\s+shutdown'])
        # Remove the management, Loopback, and AppGig interfaces from
        # the intf_for_test list
        intf_for_test[:] = [y for y in intf_for_test if not (
            y.re_match_typed(
                r'^interface\s+(\S.*)$') == "GigabitEthernet0/0" or
            y.re_match_typed(
                r'^interface\s+(\S.*)$').startswith("Loopback") or
            y.re_match_typed(r'^interface\s+(\S.*)$').startswith("AppGig"))]
        # Test the remaining interfaces for shutdown and add to shut list
        for intf_obj in intf_for_test:
            shut_interfaces.append(
                intf_obj.re_match_typed(r'^interface\s+(\S.+?)$'))

        # Select all interfaces
        intf = parse.find_objects(r'^interface')
        if debug:
            print(f"intf = {intf}")
        # Remove the management, Loopback, and AppGig interfaces from
        # the interface list
        intf[:] = [x for x in intf if not (
            x.re_match_typed(
                r'^interface\s+(\S.*)$') == "GigabitEthernet0/0" or
            x.re_match_typed(
                r'^interface\s+(\S.*)$').startswith("Loopback") or
            x.re_match_typed(r'^interface\s+(\S.*)$').startswith("AppGig"))]

        for intf_obj in intf:
            # Get the interface name
            intf_name = intf_obj.re_match_typed(r'^interface\s+(\S.*)$')
            # Default to include this interface_descriptor
            intf_include = True

            # Only interface name will be used to catogrize different types of
            # interfaces (downlink and uplink)
            only_intf_name = re.sub("\d+|\\/", "", intf_name)  # DON'T EDIT!
            if not only_intf_name == "Vlan":
                if only_intf_name == "Port-channel":
                    Switch_module = "0"
                else:
                    Switch_module = intf_obj.re_match_typed(
                        r'^interface\s\S+?thernet+(\d)')
                    if Switch_module == "":
                        Switch_module = intf_obj.re_match_typed(
                            r'^interface\s\S+?GigE+(\d)')
                port, sub_module = check(intf_name)
                if sub_module == "1":
                    intf_include = False
                    if debug:
                        print(f"Switch_module = {Switch_module}")
                        print(f"nm_list = {nm_list}")
                    if not nm_list[int(Switch_module)-1] == "":
                        if nm_dict[
                              nm_list[int(Switch_module)-1]]['supported']:
                            for regex in nm_dict[
                              nm_list[int(Switch_module)-1]]['ports']:
                                if re.match(regex,intf_name):
                                    # Make sure it has some settings,
                                    # or skip it
                                    if not intf_obj.children == []:
                                        intf_include = True
                                        if debug:
                                            print(f"I matched {intf_name} "+
                                                  f"to {regex}\n")
                                        break
                if intf_include:
                    # If we are including the interfaces, let's set a few
                    # default settings
                    All_interfaces[only_intf_name].append(intf_name)
                    port_dict[intf_name] = {}
                    port_dict[intf_name]['sw_module'] = "1"
                    port_dict[intf_name]['sub_module'] = sub_module
                    port_dict[intf_name]['port'] = port
                    port_dict[intf_name]['mac'] = []
                    port_dict[intf_name]['active'] = "true"

                    if debug:
                        print("Made it to Switch_module test")
                    if Switch_module == "0":
                        port_dict[intf_name]['sw_module'] = "1"
                    if not Switch_module == "" and not Switch_module == "0":
                        port_dict[intf_name]['sw_module'] = Switch_module

                    port_dict[intf_name]['mac'].clear()
                    # Check if the interface in the shutdown list then
                    # mark it as shutdown
                    if intf_name in shut_interfaces:
                        port_dict[intf_name]['active'] = "false"

            else:
                # interface Vlan
                All_interfaces[only_intf_name].append(intf_name)
                port_dict[intf_name] = {}
                port_dict[intf_name]['active'] = "true"
                port_dict[intf_name]['vlan'] = intf_obj.re_match_typed(
                    r'^interface\sVlan(\d+)')
                if debug:
                    print(f"for {intf_name}, vlan is " +
                          f"{port_dict[intf_name]['vlan']}")

            if intf_include:
                if debug:
                    print("Checking children")
                int_fx = intf_obj.children
                if debug:
                    print(f"Children are: {int_fx}")

                # Capture the configuration of the interface
                port_sec_raw = max_mac = ""
                r = \
                  r'\sswitchport\sport-security\smac-address\ssticky\s+(\S.+)'
                for child in int_fx:
                    # Try out our mc_pedia
                    if debug:
                        print(f"child = {child}")
                    for key, val in mc_pedia['port'].items():
                        newvals = {}
                        if debug:
                            print(f"key, val = {key},{val}")
                        if not val['regex'] == "":
                            if not child.re_match_typed(
                              regex=val['regex']) == "":
                                if not val['iosxe'] == "":
                                    exec(val.get('iosxe'), locals(), newvals)
                                    if debug:
                                        print(f"newvals[{key}] = " +
                                              f"{newvals[key]}")
                                    if not newvals[key] == "":
                                        i = intf_name
                                        port_dict[i][key] = newvals[key]
                    try:
                        port_sec_raw = child.re_match_typed(regex=r)
                    except:
                        pass
                    try:
                        regex = r'\sswitchport\sport-security\smaximum\s+(\d)'
                        max_mac = child.re_match_typed(regex=regex)
                    except:
                        pass
                    if not port_sec_raw == "":
                        port_dict[intf_name]['mac'].append(
                          mac_build(port_sec_raw))
                    if not max_mac == "":
                        port_dict[intf_name]['Port_Sec'] = max_mac

        Intf_list, Other_list = split_down_up_link(
            All_interfaces,
            Gig_uplink)
        if debug:
            print(f"Intf_list = {Intf_list}\n")
            print(f"Other_list = {Other_list}\n")
            print(f"port_dict = {port_dict}\n")
            print(f"switch_dict = {switch_dict}\n")
        return Intf_list, Other_list, port_dict, switch_dict

    def split_down_up_link(interfaces_list, Gig_uplink):
        """
        This sub-function takes a list of interfaces and sorts them into
        uplinks, downlinks and others.
        :param interfaces_list: The dictlist of interfaces categorized by type
        :param Gig_uplink:
        :return: Lists of uplinks, downlinks and other ports
        """
        Intf_list = list()
        Other_list = list()

        # Creating a copy of the interface list to avoid Runtime error
        interfaces_list_copy = interfaces_list.copy()
        if debug:
            print(f"interfaces_list = {interfaces_list}\n")
        for key, value in interfaces_list.items():
            for value in interfaces_list_copy[key]:
                if key == "HundredGigabitEthernet" or \
                  key == "FortyGigabitEthernet" or \
                  key == "TwentyFiveGigE" or \
                  key == "TenGigabitEthernet" or \
                  key == "GigabitEthernet" or \
                  key == "FastEthernet" or \
                  key == "Vlan":
                    # pass
                    Intf_list.append(value)
                else:
                    Other_list.append(value)
        return Intf_list, Other_list

    # rebuild the mac address to match Meraki format
    def mac_build(my_str, group=2, char=':'):
        """
        This sub-function takes an IOSXE MAC address and reformats it into the
        Meraki format.
        :param my_str: The MAC address from IOSXE
        :param group: How many digits to group between delimiters (2 = :00:)
        :param char: The delimiter to use between digit groupings
        :return: A string containing the translate MAC address
        """
        port_sec = re.findall(re.compile(r'[a-fA-F0-9.]{14}'), my_str)[0]
        new_p = re.sub("\.", "", port_sec)  # DO NOT EDIT THIS FOR FLAKE8!
        my_str = str(new_p)

        last = char.join(my_str[i:i+group]
                         for i in range(0, len(my_str), group))
        return last

    # Extract out the details of the switch module and the port number
    def check(intf):
        """
        This sub-function takes an IOSXE long interface name and splits it
        into a submodule and a port.
        :param intf: An IOSXE long interface name
        :return: String with the submodule number and one with the port number
        """
        if debug:
            print(f"Checking interface {intf}.")
        obj = re.search(
          r'(?:Ethernet|GigE|channel)(\d+)\/(\d+)\/(\d+)$', intf)
        if obj is not None:
            if debug:
                print(f"obj.group(0) = {obj.group(0)}")
                print(f"obj.group(1) = {obj.group(1)}")
                print(f"obj.group(2) = {obj.group(2)}")
                print(f"obj.group(3) = {obj.group(3)}")
            port = obj.group(3)
            Sub_module = obj.group(2)
        else:
            obj = re.search(r'(?:channel)(\d+)$', intf)
            if debug:
                print(f"obj.group(0) = {obj.group(0)}")
                print(f"obj.group(1) = {obj.group(1)}")
            port = obj.group(1)
            Sub_module = 0
        return port, Sub_module

    Interfaces, Others, port_dict, switch_name = read_Cisco_SW()
    if debug:
        print(f"\nInterfaces = {Interfaces}")
        print(f"\nOthers = {Others}")
    return Interfaces, Others, port_dict, switch_dict


def MerakiConfig(dashboard, organization_id, switch_path, sw_list, port_dict,
                  Intf_list, Other_list, switch_dict, nm_list):
    """
    This parent function will convert Catalyst switch config features to
    Meraki features and send them to Dashboard to program Meraki switches.
    :param dashboard: Active Meraki dashboard API session to use
    :param organization_id: Meraki Org ID used in batch configuration
    :param sw_list: List of Meraki switch serial numbers to configure
    :param port_dict: Dictionary of IOSXE ports and features from Evaluate
    :param Intf_list: List of standard ports & L3 interfaces to configure
    :param Other_list: List of other ports to configure (Port-channels)
    :param switch_name: The catalyst IOSXE hostname
    :param nm_list: List of Network Modules per switch
    :return: Lists of configured and unconfigured ports, and a list with a URL
    :      :  to each of the Meraki switches that we configured or modified
    """

    debug = DEBUG or DEBUG_TRANSLATOR

    # Create a batch action lists
    action_list = list()
    all_actions = list()
    returns_dict = {}
    post_ports_list = list()
    # Create good and bad port lists
    conf_ports = defaultdict(list)
    unconf_ports = defaultdict(list)

    if debug:
        print(f"Other_list = {Other_list}")
        for port in Other_list:
            if port in port_dict.keys():
                print(f"port_dict['{port}'] = {port_dict[port]}")

    # Create a place to hold all of the arguments to send to Dashboard
    # to update a switch port
    args = list(dict())

    # Loop to go through all the ports of the switches
    def loop_configure_meraki(port_dict, Intf_list, switch_dict):
        """
        This sub-function does the actual work of setting the meraki functions
        based on the dictionary of IOSXE ports and features from Evaluate.
        :param port_dict: Dictionary of IOSXE ports and features from Evaluate
        :param Intf_list: The list of ports to configure
        :param switch_name: The catalyst IOSXE hostname
        :return: NONE - modifies global variables in the parent function
        """

        # Create a place to hold the Dashboard URL for each switch
        # Configure the switch_name in the Dashboard
        if debug:
            print(f"switch_dict = {switch_dict}")
        for key, val in mc_pedia['switch'].items():
            if (val['translatable'] == "✓" and
                val['meraki']['skip'] == "post_process") or \
                 val['meraki']['skip'] == "post_ports":
                if debug:
                    print(f"key = {key}, val = {val}")
                if val['meraki']['skip'] == "post_ports":
                    post_ports_list.append([key, True])
                newvals = {}
                exec(val['meraki'].get('post_process'), locals(), newvals)
                if debug:
                    print(f"newvals = {newvals}")
                return_vals = newvals['return_vals']
                if debug:
                    print(f"newvals['return_vals'] = " +
                          f"{newvals['return_vals']}")
                n = 0
                while n < len(return_vals):
                    switch_dict[return_vals[n]] = newvals[return_vals[n]]
                    returns_dict[return_vals[n]] = newvals[return_vals[n]]
                    n += 1
                if debug:
                    print(f"switch_dict = {switch_dict}")

        # Loop to get all the interfaces in the port_dict
        y = 0
        while y <= len(Intf_list)-1:
            interface_descriptor = Intf_list[y]
            intf_settings = port_dict[interface_descriptor]
            if debug:
                print("\n----------- "+interface_descriptor+" -----------")
                pprint.pprint(intf_settings)

            # Check the switch that mapped to those catalyst ports
            try:
                switch_num = int(intf_settings['sw_module'])
            except:
                switch_num = 1
            if not switch_num == 0:
                switch_num -= 1

            if 'Vlan' not in interface_descriptor:
                # Setup the features for a physical interface
                if intf_settings['sub_module'] == "1":
                    if not nm_list[switch_num] == "":
                        for regex in nm_dict[nm_list[switch_num]]['ports']:
                            if re.match(regex,interface_descriptor):
                                if debug:
                                    print(f"I matched {interface_descriptor}"+
                                          f" to {regex}\n")
                                port = "1_"+nm_list[switch_num]
                                port += "_"+intf_settings['port']
                                args.append([sw_list[switch_num],
                                             port, {}])
                else:
                    args.append([sw_list[switch_num],
                                 intf_settings['port'], {}])
                # Setup the default features
                args[y][2].update({
                    'enabled': True,
                    'tags': [],
                    'poeEnabled': True,
                    'isolationEnable': False,
                    'rstpEnabled': True,
                    'accessPolicyType': 'Open'})
                for key, val in mc_pedia['port'].items():
                    newvals = {}
                    if val['meraki']['skip'] is not True:
                        # Apply any post processing for Meraki config
                        if val['meraki']['skip'] in [
                                                     'post_process',
                                                     'post_ports']:
                            exec(val['meraki'].get('post_process'),
                                                   locals(),
                                                   newvals)
                            if debug:
                                print(f"newvals = {newvals}")
                            if val['meraki']['skip'] == 'post_process':
                                if not newvals[key] == "":
                                    # Apply post processing port setting
                                    # returned for that key
                                    intf_settings[key] = newvals[key]
                        try:
                            # Update the features we will later apply
                            # for this interface
                            args[y][2].update({key: intf_settings[key]})
                        except:
                            if "default" in val['meraki']:
                                # We weren't given a value for this feature,
                                # apply the Meraki default value
                                intf_settings[key] = val['meraki']['default']
                                args[y][2].update({key: intf_settings[key]})
                        if "return_vals" in newvals:
                            # We must have called a post-process feature in
                            # the encyclopedia that also has a
                            # post-port-process, so let's keep the extra
                            # return values for that extra process
                            return_vals = newvals['return_vals']
                            if debug:
                                print("newvals['return_vals'] = " +
                                      f"{newvals['return_vals']}")
                            n = 0
                            while n < len(return_vals):
                                if debug:
                                    print(f"return_vals[{n}] = " +
                                          f"{return_vals[n]}")
                                    print("post_ports_list = " +
                                          f"{post_ports_list}")
                                # Data returned for post_ports_processing
                                # looks like:
                                # encyclopedia_key, [a list of values]
                                # There can be multiple instances, so we will
                                # append them all to a list of lists
                                post_ports_list.append(
                                    [return_vals[n],
                                     newvals[return_vals[n]]])
                                if debug:
                                    print("post_ports_list = " +
                                          f"{post_ports_list}")
                                n += 1
                try:
                    # If port was disabled, disable it in the port)_dict
                    args[y][2].update({'enabled': False
                                       if intf_settings["active"] == "false"
                                       else True})
                except:
                    pass
                # Check if the interface mode is configured as Access
                if intf_settings['type'] == "access":
                    # For access ports, apply port security and sticky MACs
                    # as needed
                    try:
                        if not intf_settings['mac'] == []:
                            pass
                    except:
                        pass
                    try:
                        if not intf_settings['Port_Sec'] == "":
                            mac_limit = intf_settings['Port_Sec']
                    except:
                        intf_settings['Port_Sec'] = ""
                    if not intf_settings['Port_Sec'] == "":
                        args[y][2].update({
                            'accessPolicyType': 'Sticky MAC allow list',
                            'stickyMacAllowList': json.dumps(
                              intf_settings['mac']),
                            'stickyMacAllowListLimit': mac_limit})

            else:
                # Setup the features for a logical L3 interface
                # Setup the default features
                args.append([returns_dict['networkId'],
                            returns_dict['switchStackId'],
                            interface_descriptor,
                            intf_settings['vlan'],
                            {}])
                for key, val in mc_pedia['layer3'].items():
                    newvals = {}
                    if val['meraki']['skip'] is not True:
                        # Apply any post processing for Meraki config
                        if val['meraki']['skip'] in ['post_process',
                                                     'post_ports']:
                            exec(val['meraki'].get('post_process'),
                                                   locals(),
                                                   newvals)
                            if debug:
                                print(f"newvals = {newvals}")
                            if val['meraki']['skip'] == 'post_process':
                                if not newvals[key] == "":
                                    # Apply post processing port setting
                                    # returned for that feature
                                    intf_settings[key] = newvals[key]
                        if debug:
                            print(f"key = {key}, " +
                                  f"newvals[key] = {newvals[key]}")
                        try:
                            # Update the features we will later apply
                            # for this interface
                            args[y][4].update({key: intf_settings[key]})
                        except:
                            if "default" in val['meraki']:
                                # We weren't given a value for this feature,
                                # apply the Meraki default value
                                intf_settings[key] = val['meraki']['default']
                                args[y][4].update({key: intf_settings[key]})
                        if "return_vals" in newvals:
                            # We must have called a post_process feature in
                            # the encyclopedia that also has a
                            # post_port_process, so let's keep the extra
                            # return values for that extra process
                            return_vals = newvals['return_vals']
                            if debug:
                                print("newvals['return_vals'] = " +
                                      f"{newvals['return_vals']}")
                            # On L3 interfaces, save the extra return values
                            # in the port_dict for that port for
                            # post_ports processing
                            n = 0
                            while n < len(return_vals):
                                if debug:
                                    print(f"return_vals[{n}] = " +
                                          f"{return_vals[n]}")
                                args[y][4].update(
                                  {return_vals[n]: newvals[return_vals[n]]})
                                if debug:
                                    print(f"args for {interface_descriptor}" +
                                          f" = {args[y][4]}")
                                n += 1
                if debug:
                    print(f"args[{y}] = {args[y]}")
            # Append args to the port_dict as meraki_args
            port_dict[interface_descriptor]['meraki_args'] = args[y]

            # Append the port update call to Dashboard to the batch list
            if debug:
                print("Number of sub lists in action_list is " +
                      f"{len(action_list)}")
                try:
                    print("Number of batch actions in action_list" +
                          f"[{switch_num}] is {len(action_list[switch_num])}")
                except:
                    pass

            if not len(action_list) == switch_num+1:
                # We are on to the next switch, so I want a new sublist
                action_list.append([])
            # If it's a physical interface,
            # Add this action to the action_list sublist for the switch
            if 'Vlan' not in interface_descriptor:
                try:
                    action_list[switch_num].append(
                        dashboard.batch.switch.updateDeviceSwitchPort(
                            args[y][0],
                            args[y][1],
                            **args[y][2]))
                except:
                    unconf_ports[switch_num].append(args[y][1])
                    print("We caught an exception configuring " +
                          f"{args[y][1]} on {switch_num}")
                if not args[y][1] in unconf_ports[switch_num]:
                    conf_ports[switch_num].append(args[y][1])
            y += 1

        # post_ports processing
        if debug:
            print(f"At post_ports, port_dict = {port_dict}")
            print(f"Length of post_ports_list = {len(post_ports_list)}")
            print(f"post_ports_list = {post_ports_list}")
        if not len(post_ports_list) == 0:
            # In case there are multiple entries for the same
            # post_ports_process, create a consolidated list so that we only
            # call each one once
            short_list = list(set(map(lambda x: x[0], post_ports_list)))
            if debug:
                print(f"short_list = {short_list}")

            # This is the processing loop for port-level post ports processes
            item = 0
            while item < len(short_list):
                if debug:
                    print(f"short_list[{item}] = {short_list[item]}")
                # Call the post-port process for that feature if
                # there is one in the PORTS section of the encyclopedia
                newvals = {}
                if short_list[item] in mc_pedia['port'].keys():
                    exec(mc_pedia['port'][short_list[item]]['meraki'].get(
                      'post_ports_process'), locals(), newvals)
                    if "return_vals" in newvals:
                        return_vals = newvals['return_vals']
                        if "channel_port_dict" in return_vals:
                            if debug:
                                print("newvals['channel_port_dict'] = " +
                                      f"{newvals['channel_port_dict']}")
                            for a_port in newvals['channel_port_dict']:
                                if debug:
                                    print(f"a_port = {a_port}")
                                port_dict[a_port] = newvals[
                                    'channel_port_dict'][a_port]
                                if debug:
                                    print("port_dict[a_port] = " +
                                          f"{port_dict[a_port]}")
                        if 'conf_ports' in return_vals:
                            cp_list = newvals['conf_ports']
                            if debug:
                                print(f"cp_list = {cp_list}")
                            for port in cp_list:
                                switch_num = "stack" if len(
                                    sw_list) > 1 else 1
                                conf_ports[switch_num].append(port)
                        if 'unconf_ports' in return_vals:
                            up_list = newvals['unconf_ports']
                            if debug:
                                print(f"up_list = {up_list}")
                            for port in up_list:
                                switch_num = "stack" if len(
                                    sw_list) > 1 else 1
                                unconf_ports[switch_num].append(port)
                item += 1

            # This is not used yet, but I left this here in case we need
            # it later for switch-level post ports processes
            item = 0
            while item < len(short_list):
                if debug:
                    print(f"short_list[{item}] = {short_list[item]}")
                # Call the post-port process for that feature if
                # there is one in the SWITCH section of the encyclopedia
                newvals = {}
                if short_list[item] in mc_pedia['switch'].keys():
                    exec(mc_pedia['switch'][short_list[item]]['meraki'].get(
                        'post_ports_process'), locals(), newvals)
                    if debug:
                        print(f"return_vals = {return_vals}")
                    if "return_vals" in newvals:
                        return_vals = newvals['return_vals']
                        pass
                item += 1

        if debug:
            print(f"\n\naction_list = {action_list}\n\n")
    loop_configure_meraki(port_dict, Intf_list, switch_dict)

    # Combine all of the action_list sublists into a larger set for batching
    x = 0
    while x <= len(action_list)-1:
        all_actions.extend(action_list[x])
        x += 1
    if debug:
        print(f"all_actions = {all_actions}")

    # Save an action batch file for the port features
    # dir = os.path.join(os.getcwd(),"../../files")
    # with open(os.path.join(dir,switch_path+".ab0"), 'w') as file:
    #    file.write(json.dumps(all_actions)) # use `json.loads` to reverse
    #    file.close()

    if debug:
        print(f"Number of batch actions for Dashboard: {len(action_list)}\n")
        print(f"Args listdict is: {args}\n")

    test_helper = batch_helper.BatchHelper(dashboard,
                                           organization_id,
                                           all_actions,
                                           linear_new_batches=False,
                                           actions_per_new_batch=100)
    test_helper.prepare()
    # test_helper.generate_preview()
    test_helper.execute()
    if debug:
        print(f'helper status is {test_helper.status}')

    # try:
    batches_report = dashboard.organizations.getOrganizationActionBatches(
        organization_id)
    # except:
    #    pass
    new_batches_statuses = [{'id': batch['id'], 'status': batch['status']}
                            for batch in batches_report if batch['id']
                            in test_helper.submitted_new_batches_ids]
    if debug:
        print(f'Batch status returned is: {new_batches_statuses}')
    failed_batch_ids = [batch['id'] for batch in new_batches_statuses
                        if batch['status']['failed']]
    if debug:
        print(f'Failed batch IDs are as follows: {failed_batch_ids}')

    if debug:
        print(f"\nport_dict = {port_dict}\n")

    return conf_ports, unconf_ports, port_dict, \
        returns_dict['urls'], returns_dict['networkId']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md') as history_file:
    history = history_file.read()

requirements = [
    "webexteamsbot==0.1.4.2",
    "webexteamssdk==1.0.3",
    "Flask>=0.12.1",
    "netmiko==4.3.0",
    "tabulate==0.9.0",
    "ciscoconfparse2==0.5",
    "meraki==1.42.0",
    "ngrok==1.1.0",
    "python-docx==1.1.0",
    "docx2pdf==0.1.8",
    "requests==2.31.0"
    ]

setup_requirements = [ ]

setup(
    author="Ed Coen",
    author_email='ecoen@cisco.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3 :: Only',
    ],
    description="An app to check and translate Catalyst switch configs to Meraki.",
    install_requires=requirements,
    license="MIT license",
    long_description_content_type="text/markdown",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='merakicat,catalyst,meraki,cisco,migration,webexteamsbot',
    name='merakicat',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    setup_requires=setup_requirements,
    url='https://github.com/ecoen66/merakicat',
    zip_safe=False,
)



# Batch helper error
class BatchHelperError(Exception):
    """
    Base class for other exceptions.

    Attributes:
        message -- explanation of the error
    """

    message = "Something went wrong."

    def __init__(self, message=message):
        self.message = message
        super().__init__(self.message)


# To catch exceptions when using the batch helper
class IntervalFactorTooSmallError(BatchHelperError):
    """
    Raised when the polling interval is too short.

    Attributes:
        interval_factor -- the requested interval factor
        message -- explanation of the error
    """

    def __init__(self, interval_factor):
        self.interval_factor = interval_factor
        self.message = f'The polling interval factor must be at least {MINIMUM_INTERVAL_FACTOR}.'
        super().__init__(self.message)


class NotEnoughActionsError(BatchHelperError):
    """
    Raised when the number of actions is too few.

    Attributes:
        actions_per_batch -- the requested number of actions per batch
        message -- explanation of the error
    """

    def __init__(self, actions_per_batch):
        self.actions_per_batch = actions_per_batch
        self.message = f'The minimum number of actions in a batch is {MIN_ACTIONS}.'
        super().__init__(self.message)


class TooManyActionsError(BatchHelperError):
    """
    Raised when the number of actions is too many.

    Attributes:
        actions_per_batch -- the requested number of actions per batch
        message -- explanation of the error
    """

    def __init__(self, actions_per_batch):
        self.actions_per_batch = actions_per_batch
        self.message = f'The maximum number of actions in a batch is {MAX_ACTIONS_ASYNC}.'
        super().__init__(self.message)


class TooManySynchronousActionsError(BatchHelperError):
    """
    Raised when the number of synchronous actions is too many.

    Attributes:
        actions_per_batch -- the requested number of actions per batch
        message -- explanation of the error
    """

    def __init__(self, actions_per_batch):
        self.actions_per_batch = actions_per_batch
        self.message = f'The maximum number of actions in a synchronous batch is {MAX_ACTIONS_SYNC}.'
        super().__init__(self.message)


class RequiredBatchIdError(BatchHelperError):
    """
    Raised when the user doesn't provide a batch ID.

    Attributes:
        actions_per_batch -- the requested number of actions per batch
        message -- explanation of the error
    """

    def __init__(self):
        self.message = f'If dependent, then you must specify depends_on_batch_id.'
        super().__init__(self.message)


class RequiredBatchOrganizationError(BatchHelperError):
    """
    Raised when the user doesn't provide a batch ID.

    Attributes:
        actions_per_batch -- the requested number of actions per batch
        message -- explanation of the error
    """

    def __init__(self):
        self.message = f'If you set same_organization False, then you must specify the alternateOrganizationId.'
        super().__init__(self.message)


class RequiredBatchFailureError(BatchHelperError):
    """
    Raised when the required batch has failed.

    Attributes:
        depends_on_batch_id -- the ID of the batch dependency
        message -- explanation of the error
    """

    def __init__(self, depends_on_batch_id, batch_errors: list = []):
        self.depends_on_batch_id = depends_on_batch_id
        self.batch_errors = batch_errors
        self.message = f'The required batch with ID {self.depends_on_batch_id} failed. The errors are {batch_errors}.'
        super().__init__(self.message)


class RequiredBatchNotFoundError(BatchHelperError):
    """
    Raised when the required batch is not found.

    Attributes:
        depends_on_batch_id -- the ID of the batch dependency
        organizationId -- the ID of the org where the batch should be
        message -- explanation of the error
    """

    def __init__(self, depends_on_batch_id, organizationId):
        self.depends_on_batch_id = depends_on_batch_id
        self.organizationId = organizationId
        self.message = f'The required batch with ID {self.depends_on_batch_id} in org ID {self.organizationId} was' \
                       f' not found.'
        super().__init__(self.message)


class RequiredBatchNotStartedError(BatchHelperError):
    """
    Raised when the required batch is not found.

    Attributes:
        depends_on_batch_id -- the ID of the batch dependency
        organizationId -- the ID of the org where the batch should be
        message -- explanation of the error
    """

    def __init__(self, depends_on_batch_id, organizationId):
        self.depends_on_batch_id = depends_on_batch_id
        self.organizationId = organizationId
        self.message = f'The required batch with ID {self.depends_on_batch_id} in org ID {self.organizationId} has' \
                       f' not been started (e.g. set confirmed to true). Start any required batches before using this' \
                       f' feature!'
        super().__init__(self.message)


class RequiredBatchStillInProgress(BatchHelperError):
    """
    Raised when the required batch is not found.

    Attributes:
        depends_on_batch_id -- the ID of the batch dependency
        organizationId -- the ID of the org where the batch should be
        message -- explanation of the error
    """

    def __init__(self, depends_on_batch_id, organizationId):
        self.depends_on_batch_id = depends_on_batch_id
        self.organizationId = organizationId
        self.message = f'The required batch with ID {self.depends_on_batch_id} in org ID {self.organizationId} did' \
                       f' not complete within the maximum wait time.'
        super().__init__(self.message)


class BatchCreationFailureError(BatchHelperError):
    """
    Raised when the required batch is not found.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self):
        self.message = f'There was an error submitting an action batch.'
        super().__init__(self.message)


class LinearBatchRequirementsError(BatchHelperError):
    """
    Raised when attempting to combine unconfirmed new batches with linear new batches.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self):
        self.message = f'Linear new batches, which submits new batches in order upon completion of the prior batch,' \
                       f'requires new batches to be confirmed, so that they start immediately upon submission.'


class UnpreparedError(BatchHelperError):
    """
    Raised when attempting to execute a batch helper without preparing it first.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self):
        self.message = f'Executing a batch helper requires preparing it first.'



import subprocess
import platform


def Ping(host):
    """
    Returns True if host (str) responds to a ping request.
    Remember that a host may not respond to a ping (ICMP)
    request even if the host name is valid.
    :param host: The host name, FQDN or IP address to ping
    :return: True if reachable, False if not
    """

    # Option for the number of packets as a function of
    param = '-n' if platform.system().lower() == 'windows' else '-c'

    # Building the command. Ex: "ping -c 1 google.com"
    command = ['ping', param, '1', '-q', host]

    return subprocess.call(command, stdout=subprocess.DEVNULL) == 0



from ciscoconfparse2 import CiscoConfParse
from mc_user_info import DEBUG, DEBUG_CHECKER
from mc_pedia import mc_pedia


def CheckFeatures(sw_file):
    """
    This function will check a Catalyst switch config file for feature
    mapping to Meraki.
    :param sw_file: The incoming config filespec
    :return: The hostname, supportable features, unsupported features
    :      : with more info.
    """

    debug = DEBUG or DEBUG_CHECKER

    Features_configured = list()
    aux_features_config = list()
    host_name = ""

    # Here we will go through parsing/reading Cisco Catalyst configuration
    # file and capture specific configuration
    parse = CiscoConfParse(sw_file, syntax='ios', factory=True)

    for key, val in mc_pedia['switch'].items():
        newvals = {}
        if debug:
            print(f"key,val = {key},{val}\n")
        if not val['regex'] == "":
            sample = parse.find_objects(val['regex'])
            if not sample == []:
                if debug:
                    print(f"sample = {sample}")
                exec(val.get('iosxe'), locals(), newvals)
                if debug:
                    print(f"newvals = {newvals}\n")
                if not newvals[key] == "":
                    if debug:
                        print(f"newvals[{key}] = {newvals[key]}\n")
                    if key == "switch_name":
                        host_name = newvals['host_name']
                    hold_me = list()
                    hold_me.extend([val['name'],
                                    val['support'],
                                    val['translatable']])
                    if debug:
                        print(f"hold_me = {hold_me}")
                    try:
                        hold_me.extend([val['note']])
                    except:
                        hold_me.extend([""])
                    if debug:
                        print(f"hold_me = {hold_me}")
                    try:
                        hold_me.extend([val['url']])
                    except:
                        hold_me.extend([""])
                    if debug:
                        print(f"hold_me = {hold_me}")
                    sub_feature_iterations = list()
                    for sub_sample in sample:
                        sub_feature_iterations.append([sub_sample])
                    hold_me.extend([sub_feature_iterations])
                    if debug:
                        print(f"hold_me = {hold_me}\n")
                    Features_configured.append(hold_me)

    port_details = parse.find_objects(r'^interface')
    for detail in port_details:
        check_features = Interface_detail(detail)
        if debug:
            print(f"======{check_features}")
        # Go through a loop to read the return of the config under the
        # interface and then add them to the main list (Features_configured)
        x = 0
        while x < len(check_features):
            Features_configured.append(check_features[x])
            x += 1
    # Only capture unique values in a new list
    for entry in Features_configured:
        if debug:
            print(f"entry = {entry}")
        if len(aux_features_config) == 0:
            aux_features_config.append(entry)
            if debug:
                print("Created the first aux entry")
        else:
            found = False
            for i, lst in enumerate(aux_features_config):
                for j, feature_name in enumerate(lst):
                    if feature_name == entry[0]:
                        found = True
                        if debug:
                            print(f"aux_features_config[{i}][5] = " +
                                  f"{aux_features_config[i][5]}")
                            print(f"entry[5] = {entry[5]}")
                        aux_features_config[i][5].extend(entry[5])
            if not found:
                aux_features_config.append(entry)
    if debug:
        print(f"aux_features_config = {aux_features_config}")
    return (host_name, aux_features_config)


def Interface_detail(interface_value):
    """
    This sub-function is used to read the interface configuration.
    :param interface_value: An interface object from CiscoConfParse.
    :return: A combined list of  all the features on the interface.
    """

    debug = DEBUG or DEBUG_CHECKER

    if debug:
        print(f"interface_value = {interface_value}\n")

    feature_list_on_interface = list()
    # Check the configuration of the interfaces
    interface_children = interface_value.children
    for child in interface_children:
        parent = child.parent
        for key, val in mc_pedia['port'].items():
            newvals = {}
            if debug:
                print(f"key,val = {key},{val}\n")
            if not val['regex'] == "":
                if not child.re_match_typed(regex=val['regex']) == "":
                    if debug:
                        print(f"child = {child}")
                    exec(val.get('iosxe'), locals(), newvals)
                    if debug:
                        print(f"newvals[{key}] = {newvals[key]}\n")
                    if not newvals[key] == "":
                        hold_sub = list()
                        hold_sub.extend([val['name'],
                                         val['support'],
                                         val['translatable']])
                        try:
                            hold_sub.extend([val['note']])
                        except:
                            hold_sub.extend([""])
                        try:
                            hold_sub.extend([val['url']])
                        except:
                            hold_sub.extend([""])
                        if debug:
                            print(f"hold_sub = {hold_sub}")
                        hold_sub.append([[child, parent]])
                        if debug:
                            print(f"hold_sub = {hold_sub}")
                        feature_list_on_interface.append(hold_sub)
        if debug:
            print("feature_list_on_interface = " +
                  f"{feature_list_on_interface}")
    if debug:
        print(f"feature_list_on_interface for {interface_value} = " +
              f"{feature_list_on_interface}\n")

    # Combine all the features on the interface together
    # in a list and send it back
    return (feature_list_on_interface)
